#!/usr/bin/python

# Copyright (C) 2005 Michael Ellerman <michael@ellerman.id.au>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import patches
import os
import sys
import string
from bzrlib.commands import Command

def parse_args(args):
    if len(args) == 2 and args[1] == '--bzr-usage':
        print '\n'
        return True
    elif len(args) == 2 and args[1] == '--bzr-help':
        print 'Shelve a patch, you can get it back later with unshelve.'
        return True
    elif len(args) == 1:
        pass
    else:
        raise Exception("Don't understand args %s" % args)

    return False

def unshelve():
    root = run_bzr('root')[0].strip()
    shelf = os.path.join(root, '.bzr-shelf')

    if not os.path.exists(shelf):
        raise Exception("No shelf found in '%s'" % shelf)

    patch = open(shelf, 'r').read()

    print >>sys.stderr, "Reapplying shelved patches"
    pipe = os.popen('patch -d %s -s -p0' % root, 'w')
    pipe.write(patch)
    pipe.flush()

    if pipe.close() is not None:
        raise Exception("Failed running patch!")

    os.remove(shelf)
    print 'Diff status is now:'
    os.system('bzr diff | diffstat')

    return True

def shelve():
    diff_lines = run_bzr('diff')
    patch_list = patches.parse_patches(diff_lines.__iter__())
    to_shelve = prompt_user(patch_list)

    if len(to_shelve) == 0:
        print >>sys.stderr, 'Nothing to shelve'
        return True

    root = run_bzr('root')[0].strip()
    shelf = os.path.join(root, '.bzr-shelf')
    print >>sys.stderr, "Saving shelved patches to", shelf
    shelf = open(shelf, 'a')

    for patch in to_shelve:
        shelf.write(str(patch))

    shelf.flush()
    os.fsync(shelf.fileno())
    shelf.close()

    print >>sys.stderr, "Reverting shelved patches"
    pipe = os.popen('patch -d %s -sR -p0' % root, 'w')
    for patch in to_shelve:
       pipe.write(str(patch))
    pipe.flush()

    if pipe.close() is not None:
        raise Exception("Failed running patch!")

    print 'Diff status is now:'
    os.system('bzr diff | diffstat')

    return True

def run_bzr(args):
    if type(args) is str:
        args = [ args ]
    pipe = os.popen('bzr %s' % string.join(args, ' '), 'r')
    lines = pipe.readlines()
    if pipe.close() is not None:
        raise Exception("Failed running bzr")
    return lines

def prompt_user(patch_list):
    total = 0
    for patch in patch_list:
        total += len(patch.hunks)

    out = sys.stdout
    inp = sys.stdin

    to_shelve = []
    i = 1
    for patch in patch_list:
        if patch.oldname != patch.newname:
            name = "%s -> %s" % (patch.oldname, patch.newname)
        else:
            name = patch.oldname

        hunks = patch.hunks[:]
        for hunk in hunks:
            print name
            print hunk
            while True:
                out.write('Shelve this change? (%d of %d) [yn] ' % (i, total))
                line = inp.readline().strip().lower()
                if line == 'y' or line == 'yes':
                    break
                elif line == 'n' or line == 'no':
                    patch.hunks.remove(hunk)
                    break

            i += 1

        if len(patch.hunks) > 0:
            to_shelve.append(patch)

    return to_shelve

class cmd_shelve(Command):
    """Temporarily remove some changes from the current tree.
    Use 'unshelve' to restore these changes.
    """

    def run(self):
        return shelve()

class cmd_unshelve(Command):
    """Restore previously-shelved changes to the current tree.
    See also 'shelve'.
    """
    def run(self):
        return unshelve()
